#ifndef __DBTABLE_HPP_
#define __DBTABLE_HPP_

#include <string>
#include <vector>
#include <algorithm>

#include "NovaDatabase/Row.h"
#include "NovaDatabase/NovaDatabase.h"

#include <libpq-fe.h>

#include <boost/thread/mutex.hpp>

namespace nova {
  namespace database {

    enum DBTableType {
      kGenericTable,
      kValidityTable,
      kHardwareTable,
      kDAQTable,
      kDCSTable,
      kNTableType
    };

    /**
     * Generalized Database Table Interface
     *
     * @author Jonathan Paley
     * @version $Id: Table.h,v 1.63 2015/02/09 21:26:43 jpaley Exp $
     */
    class Table
    {
    public:
      
      Table();
      Table(std::string tname);
      Table(const char* tname);
      ~Table();

      std::string Name() { return fTable->tableName();}
      std::string DBName() { return fDBName;}
      std::string DBHost() { return fDBHost;}
      std::string User() { return fUser; }
      std::string Role() { return fRole; }
      std::string DBPort() { return fDBPort; }
      int         TableType() { return fTableType; }

      bool SetTableType(int t);

      bool CreateFromDB(const char* schemaName, const char* tableName, 
			const char* dbhost=0, const char* dbname=0,  
			const char* dbport=0, const char* dbuser=0);
      
      void SetIgnoreEnvVar(bool f) { fIgnoreEnvVar = f;}
      void SetUser(std::string uname) { fUser = uname; }
      void SetUser(const char* uname) { fUser = uname; }
      bool SetRole(std::string role);
      bool SetRole(const char* role);
      void SetDBName(std::string dbname) { fDBName = dbname; }
      void SetDBName(const char* dbname) { fDBName = dbname; }
      void SetDBHost(std::string dbhost) { fDBHost = dbhost; }
      void SetDBHost(const char* dbhost) { fDBHost = dbhost; }
      void SetDBPort(std::string p) { fDBPort = p; }
      void SetDBPort(const char* p) { fDBPort = p; }
      void SetDBInfo(std::string name, std::string host, std::string port,
		     std::string user);
      void SetDBInfo(const char* name, const char* host, const char* port,
		     const char* user);

      bool SetPasswordFile(const char* fname=0); ///< fname should be the name of the file 
      void ClearPassword() { fPassword=""; }
      void DisablePasswordAccess() { fPassword = ""; }
 
      void ResetRole() { fRole = fUser; }

      void SetVerbosity(int i) { fVerbosity = i;}

      int NCol() {return fCol.size();}
      int NRow() {return fRow.size();}

      void Clear() { 
	fRow.clear(); fValidityStart.clear(); fValidityEnd.clear();
	fOrderCol.clear(); fDistinctCol.clear(); fNullList.clear();
	fValiditySQL = "";
	fValidityChanged = true;
      }
      void ClearRows() { fRow.clear(); fNullList.clear(); fValidityChanged=true;}
      
      Row* const GetRow(int i);
      Row* const GetRowByPKey(std::vector<std::string>& pkeyval);
      std::vector<Row*> const GetRowByColSet(std::vector<Column>& cols);

      void AddRow(const Row* row);
      void AddRow(const Row& row);

      bool RemoveRow(int i); ///< note, this will only delete a row from 
                             ///< memory, it will not delete an existing 
                             ///< row in a dB!
      
      Row* const NewRow() 
	{Row* r = new Row(fCol); return r; }

      std::vector<std::string> GetColNames();
      std::string GetColName(int i) {return fCol[i].Name(); }

      const Column* GetCol(int i) {return &fCol[i]; }
      const Column* GetCol(std::string& cname);
      const Column* GetCol(const char* cname) 
      { std::string cstr(cname); return GetCol(cstr); }

      void SetTolerance(std::string& cname, float t);
      void SetToleranceType(std::string& cname, int t);
      float Tolerance(std::string& cname);
      float ToleranceType(std::string& cname);
      
      bool ExistsInDB();
      bool ExecuteSQL(std::string cmd, PGresult*& res);

      bool LoadFromCSV(std::string fname, bool update=false);
      bool LoadFromCSV(const char* fname, bool update=false)
      { return LoadFromCSV(std::string(fname)); }
      bool UpdateFromCSV(std::string fname);
      bool UpdateFromCSV(const char* fname) 
      { return UpdateFromCSV(std::string(fname)); }
      
      bool LoadFromSSV(std::string fname, bool update=false);
      bool LoadFromSSV(const char* fname, bool update=false)
      { return LoadFromSSV(std::string(fname)); }
      bool UpdateFromSSV(std::string fname);
      bool UpdateFromSSV(const char* fname) 
      { return UpdateFromSSV(std::string(fname)); }
      
      bool LoadFromDB();
      bool WriteToDB(bool commit=true); ///< use commit=false if just testing
      bool BulkInsertInDB(int rowsPerGroup=100);
      bool WriteToCSV(std::string fname, bool appendToFile=false);
      bool WriteToCSV(const char* fname, bool appendToFile=false) 
      { return WriteToCSV(std::string(fname)); }
      
      void ClearValidity();

      bool AddDistinctColumn(unsigned int i);
      bool AddDistinctColumn(std::string col);
      bool AddDistinctColumn(const char* col) 
      { return AddDistinctColumn(std::string(col)); }
      bool RemoveDistinctColumn(unsigned int i);
      bool RemoveDistinctColumn(std::string col);
      bool RemoveDistinctColumn(const char* col) 
      { return RemoveDistinctColumn(std::string(col)); }

      bool AddOrderColumn(unsigned int i);
      bool AddOrderColumn(std::string col);
      bool AddOrderColumn(const char* col)
      { return AddOrderColumn(std::string(col)); }
      bool RemoveOrderColumn(unsigned int i);
      bool RemoveOrderColumn(std::string col);
      bool RemoveOrderColumn(const char* col)
      { return RemoveOrderColumn(std::string(col)); }

      void SetSelectLimit(int limit) { fSelectLimit=limit; }
      void SetSelectOffset(int offset) { fSelectOffset=offset; }

      void SetOrderDesc() { fDescOrder = true; }
      void SetOrderAsc() { fDescOrder = false; }

      bool GetCurrSeqVal(std::string col, int64_t& iseq);
      bool GetCurrSeqVal(const char* col, int64_t& iseq)
      { return GetCurrSeqVal(std::string(col), iseq); }

      bool HasPKey() { return fTable->hasDBPKey(); }
      int  GetNPKeyCol() { return fPKeyList.size(); }
      const Column* GetPKeyCol(int i) { return fPKeyList[i]; }

      int  GetNUnique() { return fTable->getDBUniqueCount(); }
      const dBRow_t* GetUniqueRow(int i) { 
	if (fTable->hasDBUnique() &&
	    i < this->GetNUnique())
	  return &fTable->getDBUnique(i);
	else return 0;
      }

      int  GetNCheck() { return fTable->getDBCheckCount(); }
      std::string GetCheckConstraint(int i) { 
	return std::string(fTable->getDBCheck(i).getConstraint());
      }

      void PrintPQErrorMsg() {
	if (fConnection) 
	  std::cerr << PQerrorMessage(fConnection) << std::endl;
      }

      std::string GetValiditySQL() { return fValiditySQL; }
      void SetValiditySQL(std::string cmd) { fValiditySQL = cmd; fValidityChanged=true;}

      bool SetDetector(std::string det);
      bool GetDetector(std::string& det ) const;
      bool SetDetector(int idet);
      bool GetDetector(int& idet ) const;

      void SetSchema(std::string s) { fSchema = s; }
      std::string Schema() { return fSchema; }

      template <class T>
	bool SetValidityRange(std::string cname, T start, T end) 
	{
	  Row* r = NewRow();
	  Column* c = r->Col(cname);	  
	  if (c) {	    
	    // check to see if this makes sense for a boolean column.
	    if (c->Type() == "bool") 
	      if (start != end) return false;

	    // check to see that this column wasn't already added to valid. 
	    // list if it is, overwrite it

	    unsigned int i=0;
	    for (; i<fValidityStart.size(); ++i) 
	      if (fValidityStart[i].Name() == c->Name()) break;
	    
	    if (i == fValidityStart.size()) {
	      fValidityStart.push_back(Column(*c));
	      fValidityEnd.push_back(Column(*c));
	    }
	    bool isAutoIncr = (c->Type() == "autoincr");
	    fValidityStart[i].Set(start,isAutoIncr);
	    fValidityEnd[i].Set(end,isAutoIncr);      
	    fValidityChanged=true;
	    return true;
	  }

	  delete r;
	  return false;
	}

      template <class T>
	bool SetValidityRange(std::string cname, T start) {
	return (this->SetValidityRange(cname,start,start));
      }
	
      void RemoveValidityRange(std::string& cname);
      void RemoveValidityRange(const char* cname)
      { std::string cstr(cname); return RemoveValidityRange(cstr); fValidityChanged=true;}
      
      void PrintColumns();

      bool GetConnection();
      bool CloseConnection();
      void SetMaxConnAttempts(int n) { fMaxConnAttempts=n; }
      int  GetMaxConnAttempts() { return fMaxConnAttempts; }
      bool CheckValidXMLDef();
      bool ResetConnectionInfo();

      void CacheDBCommand(std::string cmd);
      std::map<std::string,dBColumn_t*> XMLCol;

      friend std::ostream& operator<< (std::ostream& stream, const Table& t);

    private:

      void CreateFromXML(const char* fname);
      void Reset();
      bool GetConnectionInfo();

      bool CheckForNulls();

      std::vector<Column> fCol;
      std::vector<Row> fRow;
      ::std::auto_ptr<dBTable> fTable;
      bool fXMLOverDB;
      bool addInsertTime;
      bool addInsertUser;
      bool addUpdateTime;
      bool addUpdateUser;
      bool fIgnoreEnvVar;
      bool fValidityChanged;
      
      std::string fUser;
      std::string fRole;
      std::string fDBPort;
      std::string fDBHost;
      std::string fDBName;
      std::string fSchema;

      std::string fDBCacheFile;
      std::string fPassword;
      std::string fValiditySQL;

      std::vector<Column> fValidityStart;
      std::vector<Column> fValidityEnd;
      std::vector<const Column*> fPKeyList;
      std::vector<const Column*> fDistinctCol;
      std::vector<const Column*> fOrderCol;
      std::vector<std::pair<int,int> > fNullList;

      short   fVerbosity;
      bool    fDescOrder;
      bool    fIgnoreDB;
      bool    fHasValidXML;
      bool    fTestedXML;
      bool    fTestedExists;
      bool    fExistsInDB;
      int     fSelectLimit;
      int     fSelectOffset;
      int     fMaxConnAttempts;
      int     fTableType;
      bool    fHasConnection;    
      PGconn* fConnection;

      std::string fDetector;

      static boost::mutex _xsdLock;

    }; // class end

    //************************************************************

    inline std::ostream& operator<< (std::ostream& stream, const Table& t) { 
      for (unsigned int j=0; j<t.fRow.size(); ++j) {	
	stream << t.fRow[j] << std::endl;
      }      
      return stream;
    }
    
    
  } // namespace novadaqdatabase close
} // namespace nova close

#endif
