#ifndef RAWFILEPARSER_H
#define RAWFILEPARSER_H

#include <cstdlib>
#include <cstdio>
extern "C" {
#include <fcntl.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
}

#include <vector>

namespace daqdataformats {
  class RawRunHeader;
  class RawConfigurationBlock;
  class RawEvent;
}

namespace rawfileparser {

class RawFileParser {

 public:

  // Class Enums and Static data
  enum NOVA_FILE_TYPES {
    NO_FILE         = 0,
    NOVA_RUNFILE    = 1,
    NOVA_EVENTFILE,
    CORRUPT_FILE,
    UNKNOWN_FILE,
  };

  // NOVA File magic numbers
  enum NOVA_FILE_DELIMITERS {
    NOVA_ASCII   = 0x41764F4E,
    E929_ASCII   = 0x39323945,
    EVENT_MARKER = 0xE929E929,
    TAIL_MARKER  = 0x4c494154
  };

  enum EVENT_INDEX_STATE {
    NO_INDEX       = -1,
    PARTIAL_INDEX  = 0,
    FULL_INDEX     = 1
  };

  static const char* filetypes[ ];

  // Constructors

  RawFileParser();                                       //! Default constructor
  RawFileParser(const char* filename);                   //! Constructor with input filename (file is openned)

  ~RawFileParser();                                      //! Destructor

  void Init();                                           //! General initialization and reinitialization

  int open(const char* filename);                        //! Open a file for parsing
  int close();                                           //! Close any currently open file

  void setDebug(bool f){debug_flag = f;};                //! Set the debug level on or off
  void setMemoryMappedIO(bool m){memmap_io_flag = m;};   //! Set the access mode for IO calls (Memory Mapped or system read)
  bool isMemoryMappedIO(){return memmap_io_flag;};       //! Obtain the current IO access mode

  int CheckFileType();                                   //! Attempt to determine the type of file that is open
  int getFileType();                                     //! Return the type of file that is open

  // Indexing method
  int BuildEventIndex(int max_idx=-1, bool force_rebuilt = false);  //! Method that builds a full event index

  // Data Retrieval methods
  int getRunHeader(daqdataformats::RawRunHeader& theRunHeader);              //! Retrieve the run header from the file
  int getConfigBlock(daqdataformats::RawConfigurationBlock& theConfigBlock); //! Retrieve the configuration block from the file
  int getFirstEvent(daqdataformats::RawEvent& theEvent);                     //! Retrieve the first event from the file
  int getNextEvent(daqdataformats::RawEvent& theEvent);                      //! Retrieve the next event from the file
  int getEvent(daqdataformats::RawEvent& theEvent);                          //! Retrieve the next event from the file
  int getEvent(daqdataformats::RawEvent& theEvent, int event_no);            //! Retrieve the Nth event from the file
  int getPreviousEvent(daqdataformats::RawEvent& theEvent);                  //! Retrieve the previous event from the file
  int getRunTail(daqdataformats::RawRunHeader& theRunTail);                  //! Retrieve the run header from the file

  // AdvanceToXXX Function Declarations
  int AdvanceToRunHeader      (off_t& header_pos);                                 //! Moves the current parse point to the Run header
  int AdvanceToConfigStart    (off_t& conf_start_pos);                             //! Moves the current parse point to the Config Block start
  int AdvanceToFirstEvent     (off_t& first_event_pos, bool index_entries=false);  //! Moves the current parse point to the First event in the file
  int AdvanceToNextEvent      (off_t& next_event_pos,  bool index_entries=false);  //! Moves the current parse point to the Next event in the file
  int AdvanceToPreviousEvent  (off_t& next_event_pos,  bool index_entries=false);  //! Moves the current parse point to the Previous event in the file
  int AdvanceToEvent          (off_t& event_pos,       bool index_entries=false);  //! Moves to the most appropriate event in the file
  int AdvanceToConfigEnd      (off_t& conf_end_pos);                               //! Moves to the end of the configuration block
  int AdvanceToRunTail        (off_t& tail_start_pos);                             //! Moves to the start of the run tail block

  int GotoEvent(int event_no);                               //! Direct access method for events

  void*     getFileStart(){return mmfile_start;};
  uint8_t*  getFileStart8(){return mmfile_start8;};
  uint16_t* getFileStart16(){return mmfile_start16;};
  uint32_t* getFileStart32(){return mmfile_start32;};

  void*     getFileEnd()  {return mmfile_end;};
  uint8_t*  getFileEnd8() {return mmfile_end8;};
  uint16_t* getFileEnd16(){return mmfile_end16;};
  uint32_t* getFileEnd32(){return mmfile_end32;};

  void*     getCurrentBuffer()  {return mmfile_current;};    //! Return the current parse position address
  uint8_t*  getCurrentBuffer8() {return mmfile_current8;};   //! Return the current parse position address as an unsigned 8bit char*
  uint16_t* getCurrentBuffer16(){return mmfile_current16;};  //! Return the current parse position address as an unsigned 16bit int*
  uint32_t* getCurrentBuffer32(){return mmfile_current32;};  //! Return the current parse position address as an unsigned 32bit int*

  // Verification methods for buffers
  bool checkRunHeader(void* buff);      //! Check if the buffer is a valid Run Header
  bool checkConfigStart(void* buff);    //! Check if the buffer is a valid Run configuration block
  bool checkEventHeader(void* buff);    //! Check if the buffer is a valid Event Header
  bool checkEvent(void* buff);          //! Check if the buffer is a valid Event
  bool checkRunTail(void* buff);        //! Check if the buffer is a valid Run Tail

  // Buffer Utilities
  void dumpBuffer(int bytes=128, int col=4);             //! Perform a hex dump of the current position extending X bytes in N columns
  void dumpBuffer(void* buff, int bytes, int col=4);     //! Perform a hex dump of the memory pointed to by buff, extending X bytes in N columns

  int getFilefd(){return infile;};                       //! Return the raw file descriptor of the file that has been opened by the parser

  // File IO utility functions
  off_t backup(size_t bytes) {return lseek(infile,-1*bytes, SEEK_CUR);};
  off_t forward(size_t bytes){return lseek(infile,bytes, SEEK_CUR);};
  off_t rewind(off_t position){return lseek(infile, position, SEEK_SET);};
  off_t savepos(off_t& position);
  off_t curpos(){return lseek(infile, 0, SEEK_CUR);}

 protected:

  int wordsearch(unsigned int searchword);                    //! Search for the given 32bit word in the file
  int wordsearchreverse(unsigned int searchword);                    //! Search for the given 32bit word in the file

  int CheckFileType_mem();                                    //! File type checking for memory mapped IO
  int CheckFileType_file();                                   //! File type checking for raw file IO

  int checkWord(unsigned int testword);                       //! Check if the word currently pointed to in the file matches the test word
  int checkWord_mem(unsigned int testword);                   //! Memory mapped IO version
  int checkWord_file(unsigned int testword);                  //! Raw file IO version

  int GotoEvent_mem(int event_no);                               //! Direct access method for events
  int GotoEvent_file(int event_no);                               //! Direct access method for events

  int BuildEventIndex_mem(int max_idx=-1, bool force_rebuilt = false);  //! Method that builds a full event index
  int BuildEventIndex_file(int max_idx=-1, bool force_rebuilt = false);  //! Method that builds a full event index

  int AdvanceToRunHeader_mem  (off_t& header_pos);
  int AdvanceToRunHeader_file (off_t& header_pos);

  int AdvanceToConfigStart_mem  (off_t& conf_start_pos);
  int AdvanceToConfigStart_file (off_t& conf_start_pos);

  int AdvanceToFirstEvent_mem  (off_t& first_event_pos, bool index_entries=false);
  int AdvanceToFirstEvent_file (off_t& first_event_pos, bool index_entries=false);

  int AdvanceToNextEvent_mem (off_t& next_event_pos,  bool index_entries=false);
  int AdvanceToNextEvent_file(off_t& next_event_pos,  bool index_entries=false);

  int AdvanceToPreviousEvent_mem (off_t& next_event_pos,  bool index_entries=false);

  int AdvanceToRunTail_mem    (off_t& tail_start_pos);
  int AdvanceToRunTail_file   (off_t& tail_start_pos);

  ssize_t readEventSize     (bool freemem_flag=false);
  ssize_t readEventSize_mem (bool freemem_flag=false);
  ssize_t readEventSize_file(bool freemem_flag=false);

  int wordsearch_mem(unsigned int searchword);
  int wordsearch_file(unsigned int searchword);

  int wordsearchreverse_mem(unsigned int searchword);

 private:

  // Private and nimplemented to prevent copy.
  RawFileParser(RawFileParser const &other);
  RawFileParser & operator=(RawFileParser const &other);

  bool  fileopenned;   
  int   infile;
  char* infilename;
  int   filetype;
  int   fileversion;

  bool memmap_io_flag;

  // Memory Mapped IO pointers for the file

  union {
    void*     mmfile_start;
    uint8_t*  mmfile_start8;
    uint16_t* mmfile_start16;
    uint32_t* mmfile_start32;
  };

  union {
    void*     mmfile_end;
    uint8_t*  mmfile_end8;
    uint16_t* mmfile_end16;
    uint32_t* mmfile_end32;
  };

  union {
    void*     mmfile_current;
    uint8_t*  mmfile_current8;
    uint16_t* mmfile_current16;
    uint32_t* mmfile_current32;
  };

  // Memory Mapped IO pointers for the individual parts
  // of the data file
  void* header_start_addr;
  void* first_event_addr;
  void* config_block_start_addr;
  void* config_block_end_addr;
  void* tail_start_addr;

  // File offsets for old fashion read() based file IO
  off_t header_start;
  off_t first_event;
  off_t config_block_start;
  off_t config_block_end;
  off_t tail_start;

  off_t filepos;

  // Event index
  int current_event_index;
  int indexBuilt;
  int indexsize;
  std::vector <off_t> event_offset_list;
  std::vector <uint32_t*> event_addr_list;

  bool debug_flag;

};

}; // endnamespace rawfileparser
#endif  /* RAWFILEPARSER_H */
