#include "DAQDataFormats/RawNanoSliceV0.h"
#include <boost/format.hpp>

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawNanoSlice
#define  CURRENT_CLASS_VERSION        0
#define  CURRENT_FIELDS_SUB_NAMESPACE rawnano
#include "DAQDataFormats/DataFormatLast.inc"

namespace daqdataformats{

static NanoSliceVersionConvention* fNanoSliceVersionConvention = NULL;

namespace VERSION_NAMESPACE{
 //------------------------Constructors------------------------------------------
RawNanoSlice::RawNanoSlice(const version_t version):
             RawDAQData(version, true)
            ,ZEROOUT_GET_AND_SET(PulseHeight)
            ,ZEROOUT_GET_AND_SET(TimeStamp)
            ,ZEROOUT_GET_AND_SET(Reserved)
            ,ZEROOUT_FUNCTION(initWithData)
            ,ZEROOUT_FUNCTION(setParameters)
            ,ZEROOUT_FUNCTION(checkReservedForCafe)
            ,ZEROOUT_FUNCTION(initWithDataInternal)
            ,_Header(NULL)
{}
//------------------------------------------------------------------------------
RawNanoSlice::RawNanoSlice (const NanoSlice& nano):
             RawDAQData(nano.Header.Version, true)
            ,ZEROOUT_GET_AND_SET(PulseHeight)
            ,ZEROOUT_GET_AND_SET(TimeStamp)
            ,ZEROOUT_GET_AND_SET(Reserved)
            ,ZEROOUT_FUNCTION(initWithData)
            ,ZEROOUT_FUNCTION(setParameters)
            ,ZEROOUT_FUNCTION(checkReservedForCafe)
            ,ZEROOUT_FUNCTION(initWithDataInternal)
            ,_Header(NULL)
{}

//------------------------------------------------------------------------------
RawNanoSlice::RawNanoSlice (const RawNanoSlice& copyin):
             RawDAQData(copyin)
            ,DEEPCOPY_GET_AND_SET(PulseHeight)
            ,DEEPCOPY_GET_AND_SET(TimeStamp)
            ,DEEPCOPY_GET_AND_SET(Reserved)
            ,DEEPCOPY_FUNCTION(initWithData)
            ,DEEPCOPY_FUNCTION(setParameters)
            ,DEEPCOPY_FUNCTION(checkReservedForCafe)
            ,DEEPCOPY_FUNCTION(initWithDataInternal)
            ,_Header(NULL)
{
}
//------------------------Destructor--------------------------------------------
RawNanoSlice::~RawNanoSlice() {
    if(_Header)
        delete _Header;
}

//------------------------------------------------------------------------------
// Default Initialization of Data members
void RawNanoSlice::init(){

    /// After using resetBufferSource header is already constructed
    /// Now we need to initialize it
    _Header->init();

    /// Initialize the Reserved word to 0xCAFE (for version 0)
    EXECUTE_FUNCTION(setReserved)(rawnano::cafe_word);
}

//------------------------------------------------------------------------------
void RawNanoSlice::initWithData(const NanoSlice& nanoslice_parameters){

    /// After using resetBufferSource header is already constructed
    /// Now we need to initialize it
    _Header->initWithData(nanoslice_parameters.Header);

    /// Initialize the Reserved word to 0xCAFE (for version 0)
    EXECUTE_FUNCTION(setReserved)(rawnano::cafe_word);

    /// Additionally, set TimeStamp and PulseHeight from nanoslice_parameters
    EXECUTE_FUNCTION(setTimeStamp)  (nanoslice_parameters.TimeStamp);
    EXECUTE_FUNCTION(setPulseHeight)(nanoslice_parameters.PulseHeight);
}

//------------------------------------------------------------------------------
bool RawNanoSlice::checkReservedForCafe() const {return (EXECUTE_FUNCTION(getReserved)() == rawnano::cafe_word);};

//------------------------------------------------------------------------------
// Set all parameters
bool RawNanoSlice::setParameters(const NanoSlice& nanoslice_parameters){
   EXECUTE_FUNCTION(setTimeStamp)  (nanoslice_parameters.TimeStamp);
   EXECUTE_FUNCTION(setPulseHeight)(nanoslice_parameters.PulseHeight);

   /// At this point header must be already defined
   _Header->setParameters(nanoslice_parameters.Header);

   return true;
}

//------------------------------------------------------------------------------
bool RawNanoSlice::setBufferSource(const void* pointer){

   _Buffer = (void*)pointer;

   /// If Header is already defined, then set its buffer source
   /// Otherwise, need to construct it
   if(_Header)return _Header->setBufferSource(pointer);

   _Header = new daqdataformats::RawNanoSliceHeader(pointer, _version); ///< version of the NanoSliceHeader equals the version of the NanoSlice

   return true;
}

//------------------------------------------------------------------------------
// Printing the NanoSlice
void RawNanoSlice::print(std::ostream& os) const {

  _Header->print(os);

  os<<boost::format(" NanoSlice Body: TimeStamp=%u PulseHeight=%u Reserved=%u ")
                %EXECUTE_FUNCTION(getTimeStamp)()
                %EXECUTE_FUNCTION(getPulseHeight)()
                %EXECUTE_FUNCTION(getReserved)();

  const uint32_t size_of_data = EXECUTE_FUNCTION(sizeofdata)();

  os<<"Data size = "<<size_of_data<<"\n";

  
  os<<"\nRAW(hex):";
  for(uint32_t i = 0; i < size_of_data; ++i){
    printWord(i,false, os);
    os<<" ";
  }
  
  os<<"\nRAW(binary):";
  for(uint32_t i = 0; i < size_of_data; ++i){
    if(i > 0)os<<"    ";
    printWord(i,true, os);
    os<<std::endl;
  }
}


//------------------------------------------------------------------------------
uint32_t RawNanoSlice::getNSamples() const
{
    if(!fNanoSliceVersionConvention)fNanoSliceVersionConvention = new NanoSliceVersionConvention();
    return fNanoSliceVersionConvention->getNSamples(_version);
}
//------------------------------------------------------------------------------
uint32_t RawNanoSlice::getNPretriggeredSamples() const
{
    if(!fNanoSliceVersionConvention)fNanoSliceVersionConvention = new NanoSliceVersionConvention();
    return fNanoSliceVersionConvention->getNPretriggeredSamples(_version);
}
//------------------------------------------------------------------------------
Encode_Type RawNanoSlice::getEncodeType() const
{
    if(!fNanoSliceVersionConvention)fNanoSliceVersionConvention = new NanoSliceVersionConvention();
    return fNanoSliceVersionConvention->getEncodeType(_version);
}


}}// end of namespaces

