#include <DatabaseUtils/DAQConfig/GeneralDataReader.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <iostream>

namespace NOVADB = nova::database;

namespace dbutils {
namespace daqconfig {

/**
 * Constructs a data reader for the specified database table and
 * named subsystem configuration.
 */
GeneralDataReader::
GeneralDataReader(const std::string& tableDefinitionFilename,
                  const uint64_t& subsystemConfigId,
                  const std::string* namesTableDefinitionFilename)
{
    _init(tableDefinitionFilename, subsystemConfigId,
          namesTableDefinitionFilename);
}

/**
 * Constructs a data reader for the specified database table and
 * named subsystem configuration.
 */
GeneralDataReader::
GeneralDataReader(const std::string& tableDefinitionFilename,
                  const uint64_t& subsystemConfigId,
                  const std::string& primarySelectionColumnName,
                  const std::string* namesTableDefinitionFilename)
{
    _selectionColumns.push_back(primarySelectionColumnName);
    _init(tableDefinitionFilename, subsystemConfigId,
          namesTableDefinitionFilename);
}

/**
 * Constructs a data reader for the specified database table and
 * named subsystem configuration.
 */
GeneralDataReader::
GeneralDataReader(const std::string& tableDefinitionFilename,
                  const uint64_t& subsystemConfigId,
                  const std::string& primarySelectionColumnName,
                  const std::string& secondarySelectionColumnName,
                  const std::string* namesTableDefinitionFilename)
{
    _selectionColumns.push_back(primarySelectionColumnName);
    _selectionColumns.push_back(secondarySelectionColumnName);
    _init(tableDefinitionFilename, subsystemConfigId,
          namesTableDefinitionFilename);
}

/**
 * Constructs a data reader for the specified database table and
 * named subsystem configuration.
 */
GeneralDataReader::
GeneralDataReader(const std::string& tableDefinitionFilename,
                  const uint64_t& subsystemConfigId,
                  const std::string& primarySelectionColumnName,
                  const std::string& secondarySelectionColumnName,
                  const std::string& tertiarySelectionColumnName,
                  const std::string* namesTableDefinitionFilename)
{
    _selectionColumns.push_back(primarySelectionColumnName);
    _selectionColumns.push_back(secondarySelectionColumnName);
    _selectionColumns.push_back(tertiarySelectionColumnName);
    _init(tableDefinitionFilename, subsystemConfigId,
          namesTableDefinitionFilename);
}

/**
 * Constructs a data reader for the specified database table and
 * named subsystem configuration.
 */
GeneralDataReader::
GeneralDataReader(const std::string& tableDefinitionFilename,
                  const std::vector<int64_t>& subsystemConfigIdList,
                  const std::string* namesTableDefinitionFilename)
{
    _namedConfigLineage = subsystemConfigIdList;
    _init(tableDefinitionFilename, 0, namesTableDefinitionFilename);
}

/**
 * Constructs a data reader for the specified database table and
 * named subsystem configuration.
 */
GeneralDataReader::
GeneralDataReader(const std::string& tableDefinitionFilename,
                  const std::vector<int64_t>& subsystemConfigIdList,
                  const std::string& primarySelectionColumnName,
                  const std::string* namesTableDefinitionFilename)
{
    _selectionColumns.push_back(primarySelectionColumnName);
    _namedConfigLineage = subsystemConfigIdList;
    _init(tableDefinitionFilename, 0, namesTableDefinitionFilename);
}

/**
 * Constructs a data reader for the specified database table and
 * named subsystem configuration.
 */
GeneralDataReader::
GeneralDataReader(const std::string& tableDefinitionFilename,
                  const std::vector<int64_t>& subsystemConfigIdList,
                  const std::string& primarySelectionColumnName,
                  const std::string& secondarySelectionColumnName,
                  const std::string* namesTableDefinitionFilename)
{
    _selectionColumns.push_back(primarySelectionColumnName);
    _selectionColumns.push_back(secondarySelectionColumnName);
    _namedConfigLineage = subsystemConfigIdList;
    _init(tableDefinitionFilename, 0, namesTableDefinitionFilename);
}

/**
 * Constructs a data reader for the specified database table and
 * named subsystem configuration.
 */
GeneralDataReader::
GeneralDataReader(const std::string& tableDefinitionFilename,
                  const std::vector<int64_t>& subsystemConfigIdList,
                  const std::string& primarySelectionColumnName,
                  const std::string& secondarySelectionColumnName,
                  const std::string& tertiarySelectionColumnName,
                  const std::string* namesTableDefinitionFilename)
{
    _selectionColumns.push_back(primarySelectionColumnName);
    _selectionColumns.push_back(secondarySelectionColumnName);
    _selectionColumns.push_back(tertiarySelectionColumnName);
    _namedConfigLineage = subsystemConfigIdList;
    _init(tableDefinitionFilename, 0, namesTableDefinitionFilename);
}

/**
 * Handles the necessary internal initialization at construction time.
 */
void GeneralDataReader::_init(const std::string& tableDefinitionFilename,
                              const uint64_t& subsystemConfigId,
                              const std::string* namesTableDefinitionFilename)
{
    _hasValidConnection = false;
    _dbTablePtr.reset(new NOVADB::Table(tableDefinitionFilename.c_str()));
    if (_dbTablePtr->GetConnection()) {
        if (_dbTablePtr->ExistsInDB()) {
            _hasValidConnection = true;
        }
    }

    if (_namedConfigLineage.size() == 0) {
        IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        if (namesTableDefinitionFilename != 0) {
            NamedConfigUtils::
                getConfigIdAncestry(idSpec, _namedConfigLineage,
                                    *namesTableDefinitionFilename);
        }
        else {
            NamedConfigUtils::
                getConfigIdAncestry(idSpec, _namedConfigLineage);
        }
    }
}

/**
 * Does the work of setting the validity ranges in the database
 * table object based on the selection columns and values stored
 * in the data members of this class.
 */
bool GeneralDataReader::
_setValidityRanges(const std::vector<std::string>& selectionValues,
                   bool fillInDefaults)
{
    if (! _hasValidConnection) {return false;}
    for (uint32_t idx = 0; idx < _selectionColumns.size(); ++idx) {
        if (selectionValues.size() > idx) {
            if (! _dbTablePtr->SetValidityRange(_selectionColumns[idx],
                                                selectionValues[idx])) {
                return false;
            }
        }
        else if (fillInDefaults) {
            std::string defaultValue;
            DBColumnParams& params = DBColumnParams::getInstance();
            if (! params.
                columnKeyToGetDefaultValue(_selectionColumns[idx],
                                           defaultValue)) {
                return false;
            }
            if (! _dbTablePtr->SetValidityRange(_selectionColumns[idx],
                                                defaultValue)) {
                return false;
            }
        }
    }
    return true;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
