#include <NovaDAQUtilities/BackgroundProcess.h>
#include <NovaDAQUtilities/ProcessUtils.h>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include <iostream>
#include <fstream>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

namespace BFS = boost::filesystem;

namespace novadaq {

// at the moment, we don't have any mutexes protecting data that
// might be accessed in different threads since all of that data
// are simple variables, but we should look into mutexes at some point.

BackgroundProcess::BackgroundProcess(std::string commandString,
                                     std::string description) :
    _description(description) {

    //std::cout << "BP constructor, cp1 " << getpid()
    //          << ":" << this << std::endl;
    _childProcessId = -1;
    _commandString = commandString;
    _exitStatus = -99;

    _workerThread =
        new boost::thread(boost::bind(&novadaq::Runnable::run, this));
    //std::cout << "BP constructor, cp2 " << getpid()
    //          << ":" << this << std::endl;
}

BackgroundProcess::~BackgroundProcess() {

    // stop and clean up our worker thread
    //std::cout << "BP destructor, cp1 " << getpid()
    //          << ":" << this << std::endl;
    this->stop();
    //std::cout << "BP destructor, cp2 " << getpid()
    //          << ":" << this << std::endl;
    this->waitForDone();
    //std::cout << "BP destructor, cp3 " << getpid()
    //          << ":" << this << std::endl;
    delete _workerThread;
    //std::cout << "BP destructor, cp4 " << getpid()
    //          << ":" << this << std::endl;
}

// blocks!
void BackgroundProcess::waitForDone() {
    //std::cout << "BP waitForDone method, cp1 " << getpid()
    //          << ":" << this << std::endl;

    if (_hasCompleted) {return;}

    //std::cout << "BP waitForDone method, cp2 " << getpid()
    //          << ":" << this << std::endl;

    _workerThread->join();
    //std::cout << "BP waitForDone method, cp3 " << getpid()
    //          << ":" << this << std::endl;
}

/**
 * Waits until the process is done or the specified timeout
 * has expired.  Returns true if the process finishes before the
 * timeout or false if the timeout expires.
 */
bool BackgroundProcess::timedWaitForDone(uint32_t microSeconds) {

    //std::cout << "BP timedWaitForDone method " << __LINE__ << std::endl;
    if (isDone()) {return true;}

    uint32_t sleepInterval = microSeconds / 10;
    if (sleepInterval < 10) {sleepInterval = 10;}
    //std::cout << "BP timedWaitForDone method " << __LINE__ << " "
    //          << sleepInterval << std::endl;
    uint32_t loopCount = microSeconds / sleepInterval;
    if (loopCount < 1) {loopCount = 1;}
    //std::cout << "BP timedWaitForDone method " << __LINE__ << " "
    //          << loopCount << std::endl;

    for (uint32_t idx = 0; idx < loopCount; ++idx) {
        usleep(sleepInterval);
        //std::cout << "BP timedWaitForDone method " << __LINE__ << " "
        //          << isDone() << std::endl;
        if (isDone()) {return true;}
    }

    //std::cout << "BP timedWaitForDone method " << __LINE__ << std::endl;
    return false;
}

std::string BackgroundProcess::getStdoutText() {
    return _stdoutOutput;
}

std::string BackgroundProcess::getStderrText() {
    return _stderrOutput;
}

int BackgroundProcess::exitStatus() {
    return _exitStatus;
}

std::string BackgroundProcess::getFormattedSummary() {
    std::string tmpString;
    std::string workString("*** BackgroundProcess result: *** \n");
    workString.append("  Command: " + _commandString + "\n");
    workString.append("  Description: " + _description + "\n");
    workString.append("  Exit status = " + 
                      boost::lexical_cast<std::string>(_exitStatus) + "\n");

    tmpString = "  STDOUT: " + _stdoutOutput;
    boost::replace_all(tmpString, "\n", "\n  STDOUT: ");
    workString.append(tmpString + "\n");

    tmpString = "  STDERR: " + _stderrOutput;
    boost::replace_all(tmpString, "\n", "\n  STDERR: ");
    workString.append(tmpString + "\n");

    return workString;
}

void BackgroundProcess::run() {

    //std::cout << "BP run method, cp1 " << getpid()
    //          << ":" << this << std::endl;
    _running = true;

    //while (! _stopRequested) {
    //    sleep(1);
    //    std::cout << "BBB " << this
    //              << " " << _stopRequested << std::endl;
    //}

    int thisPID = getpid();
    std::string stdoutFile("/tmp/bgprocess_" +
                           boost::lexical_cast<std::string>(thisPID) +
                           "_" +
                           boost::lexical_cast<std::string>(this) +
                           ".stdout");
    std::string stderrFile("/tmp/bgprocess_" +
                           boost::lexical_cast<std::string>(thisPID) +
                           "_" +
                           boost::lexical_cast<std::string>(this) +
                           ".stderr");
    std::string retcodeFile("/tmp/bgprocess_" +
                            boost::lexical_cast<std::string>(thisPID) +
                            "_" +
                            boost::lexical_cast<std::string>(this) +
                            ".status");

    _childProcessId = fork();

    // child
    //std::cout << "BP run method, cp2 " << getpid()
    //          << ":" << this
    //          << " " << _childProcessId << std::endl;
    if (_childProcessId == 0) {
        //std::cout << "BP run method, cp3a " << getpid()
        //          << ":" << this << std::endl;

        std::string cmdStringWithRedirection("(" + _commandString + ")");
        cmdStringWithRedirection.append(" 1>" + stdoutFile);
        cmdStringWithRedirection.append(" 2>" + stderrFile);
        //std::cout << "BP run method, cp3b " << getpid()
        //          << ":" << this << " "
        //          << cmdStringWithRedirection << std::endl;

        int retcode = system(cmdStringWithRedirection.c_str());
        //std::cout << "BP run method, cp3c " << getpid()
        //          << ":" << this << " " << retcode << std::endl;

        std::ofstream statusStream(retcodeFile.c_str());
        statusStream << retcode;
        statusStream.close();
        //std::cout << "BP run method, cp3d " << getpid()
        //          << ":" << this << " " << retcode << std::endl;

        // 29-Nov-2010, KAB:  the original exit() call was taking a long
        // time to complete in some cases.  My theory is that this was
        // due to the closing of connections and other cleanup that is
        // part of the exit() operation.  So, I've added the kill -9
        // call in front of the exit() call.  This should cause the
        // child to exit without waiting (which is what we want, I believe;
        // the only purpose of the child is to run the system() call above).
        // The exit() call has been left in as a catch-all.
        kill(getpid(), 9);
        exit(retcode); // bogus
    }

    // parent
    else {
        //std::cout << "BP run method, cp4a " << getpid()
        //          << ":" << this << " " << _childProcessId << std::endl;
        int status;
        waitpid(_childProcessId, &status, 0);
        //std::cout << "BP run method, cp4b " << getpid()
        //          << ":" << this << " " << WEXITSTATUS(status) << std::endl;

        std::string tmpString;
        std::ifstream stdoutStream(stdoutFile.c_str());
        if (stdoutStream.fail()) {
            _stdoutOutput = "Error opening file " + stdoutFile;
        }
        else {
            std::getline(stdoutStream, tmpString);
            while (! stdoutStream.eof() && ! stdoutStream.fail()) {
                _stdoutOutput.append(tmpString + "\n");
                std::getline(stdoutStream, tmpString);
            }
            stdoutStream.close();

            BFS::remove(stdoutFile);
        }

        std::ifstream stderrStream(stderrFile.c_str());
        if (stderrStream.fail()) {
            _stderrOutput = "Error opening file " + stderrFile;
        }
        else {
            std::getline(stderrStream, tmpString);
            while (! stderrStream.eof() && ! stderrStream.fail()) {
                _stderrOutput.append(tmpString + "\n");
                std::getline(stderrStream, tmpString);
            }
            stderrStream.close();

            BFS::remove(stderrFile);
        }

        std::ifstream statusStream(retcodeFile.c_str());
        if (! statusStream.fail()) {
            statusStream >> _exitStatus;
            statusStream.close();
            BFS::remove(retcodeFile);
        }
    }
    //std::cout << "BP run method, cp5 " << getpid()
    //          << ":" << this
    //          << " " << _childProcessId << std::endl;

    _running = false;
    _hasCompleted = true;
}

void BackgroundProcess::stop() {
    //std::cout << "BP stop method, cp1 " << getpid()
    //          << ":" << this
    //          << " " << _childProcessId << std::endl;

    if (_hasCompleted) {return;}

    _stopRequested = true;

    if (_childProcessId > 0) {
        std::vector<int> descendantPidList;
        int status = ProcessUtils::getDescendantProcessIds(_childProcessId,
                                                           descendantPidList);
        //std::cout << "BP stop method, cp2 " << getpid()
        //          << ":" << this << " killing pid "
        //          << _childProcessId << std::endl;
        kill(_childProcessId, 9);
        if (status == 0) {
            for (std::vector<int>::size_type idx = 0;
                 idx < descendantPidList.size(); ++idx) {
                //std::cout << "BP stop method, cp3 " << getpid()
                //          << ":" << this << ", killing pid "
                //          << descendantPidList[idx] << std::endl;
                kill(descendantPidList[idx], 9);
            }
        }
        _childProcessId = 0;
    }
    //std::cout << "BP stop method, cp4 " << getpid()
    //          << ":" << this << std::endl;
}

}
