#include <NovaDAQUtilities/EnvVarCache.h>
#include <NovaDAQUtilities/CacheManager.h>
#include <boost/lexical_cast.hpp>

namespace novadaq {

/**
 * Returns the single instance of the EnvVarCache.
 *
 * @returns EnvVarCache instance.
 */
EnvVarCache& EnvVarCache::getInstance() {
    static TimedCachePolicy policy(100000);
    static EnvVarCache singletonInstance(policy);
    return singletonInstance;
}

/**
 * Default constructor (private)
 */
EnvVarCache::EnvVarCache(novadaq::CachePolicy policy) :
    novadaq::Cache(policy) {

    // register this cache with the CacheManager so that it gets
    // flushed when the CacheManager gets a flush request
    novadaq::CacheManager::getInstance().registerCache(this);
}

/**
 * Returns the value of the environmental variable with the specified name.
 * If the env var is not found, an empty string is returned.
 */
std::string EnvVarCache::getEnvVar(std::string name) {

    // protect against unsynchronized access to the cache
    boost::mutex::scoped_lock lock(_cacheMutex);

    // (global reloads are not needed since we only load values as needed)
    //_reloadIfNeeded();

    return this->_doGetEnvVar(name);
}

/**
 * Returns the value of the environmental variable with the specified name
 * with the value interpreted as a long integer.
 */
int EnvVarCache::getEnvVarAsLong(std::string name, long defaultValue) {

    // protect against unsynchronized access to the cache
    boost::mutex::scoped_lock lock(_cacheMutex);

    // (global reloads are not needed since we only load values as needed)
    //_reloadIfNeeded();

    std::string valueString = this->_doGetEnvVar(name);
    long longValue = defaultValue;
    try {
        longValue = boost::lexical_cast<long>(valueString);
    }
    catch (boost::bad_lexical_cast &excpt) {}

    return longValue;
}

/**
 * Returns the value of the environmental variable with the specified name
 * with the value interpreted as a double-precision number.
 */
double EnvVarCache::getEnvVarAsDouble(std::string name, double defaultValue) {

    // protect against unsynchronized access to the cache
    boost::mutex::scoped_lock lock(_cacheMutex);

    // (global reloads are not needed since we only load values as needed)
    //_reloadIfNeeded();

    std::string valueString = this->_doGetEnvVar(name);
    double doubleValue = defaultValue;
    try {
        doubleValue = boost::lexical_cast<double>(valueString);
    }
    catch (boost::bad_lexical_cast &excpt) {}

    return doubleValue;
}

/**
 * Returns the value of the environmental variable with the specified name,
 * for internal use.
 */
std::string EnvVarCache::_doGetEnvVar(std::string name) {

    // the higher-level methods that call this one take care
    // of the thread locking...

    // look up the requested entry in the map
    std::string value;
    std::map<std::string, std::string>::iterator iter = _envVarMap.find(name);
    if (iter == _envVarMap.end()) {

        // the getenv system call is not multi-threaded, but we don't need
        // to add explicit protection here because access to the cache is
        // protected in the public methods.
        char *envPtr = getenv(name.c_str());
        if (envPtr != 0) {
            std::string tmpString(envPtr);
            value = tmpString;
        }
        else {
            value = "";
        }
        _envVarMap[name] = value;
    }
    else {
        value = iter->second;
    }

    return value;
}

/**
 * Loads a new set of values into the cache.
 */
void EnvVarCache::_loadDataIntoCache() {

    // the higher-level methods that call this one take care
    // of the thread locking...

    // values are loaded on an as-needed basis, so there is nothing
    // to be done here
}

/**
 * Clears the data in the cache.
 */
void EnvVarCache::_clearCachedData() {

    // the higher-level methods that call this one take care
    // of the thread locking...

    _envVarMap.clear();
}

};  // end of namespace novadaq
